from typing import Type, Literal


class Argument:
    def __init__(self, arg_type: Literal['arg', 'option', 'flag', 'count'],
                 name: str, *, flag: str = None, summary: str = None,
                 value_type: Type = None, default=None, choices: list = None):
        self.arg_type = arg_type
        self.name = name
        self.flag = flag
        self.summary = summary
        self.value_type = value_type
        self.default = default
        self.choices = choices



from bolinette import blnt
from bolinette.decorators import command


@command('init_db', 'Initialize the database')
@command.argument('flag', 'seeders', flag='s', summary='Run the seeders after database creation')
async def init_db(context: 'blnt.BolinetteContext', seeders: bool):
    context.logger.info('==== Initializing database ====')
    context.logger.info('**** Dropping all tables')
    await context.db.drop_all()
    context.logger.info('**** Creating all tables')
    await context.db.create_all()
    if seeders:
        context.logger.info('**** Running seeders')
        await context.db.run_seeders(log=True, tab=5)
    context.logger.info('==== Done ====')



from bolinette.defaults.controllers.static import StaticController
from bolinette.defaults.controllers.timezone import TimezoneController
from bolinette.defaults.controllers.file import FileController
from bolinette.defaults.controllers.role import RoleController
from bolinette.defaults.controllers.user import UserController



import random
import string

import pytest

from bolinette import Bolinette
from bolinette.testing import TestClient


@pytest.fixture
def client(loop):
    async def create_client():
        return TestClient(bolinette, loop)
    bolinette = Bolinette(profile='test', overrides={
        'secret_key': ''.join(random.choices(string.ascii_letters + string.digits, k=32))
    })
    return loop.run_until_complete(create_client())



import asyncio
import inspect

from aiohttp import web as aio_web

from bolinette import blnt, console
from bolinette.blnt.commands import Parser
from bolinette.exceptions import InitError
from bolinette.utils import paths


class Bolinette:
    def __init__(self, *, profile=None, overrides=None):
        try:
            self.app = aio_web.Application()
            self.context = blnt.BolinetteContext(paths.dirname(__file__), self.app,
                                                 profile=profile, overrides=overrides)
            self.app['blnt'] = self.context
            self.run_init_functions(self.app)
        except InitError as init_error:
            console.error(f'Error raised during Bolinette init phase\n{str(init_error)}')
            exit(1)

    @staticmethod
    def run_init_functions(app):
        for func in blnt.cache.init_funcs:
            if inspect.isfunction(func):
                if inspect.iscoroutinefunction(func):
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(func(app['blnt']))
                else:
                    func(app['blnt'])

    def start_server(self, *, host: str = None, port: int = None):
        if self.context.env['build_docs']:
            self.context.docs.build()
        self.context.docs.setup()
        self.context.logger.info(f"Starting Bolinette with '{self.context.env['profile']}' environment profile")
        aio_web.run_app(self.app,
                        host=host or self.context.env['host'],
                        port=port or self.context.env['port'],
                        access_log=self.context.logger)
        self.context.logger.info(f"Bolinette stopped gracefully")

    def exec_cmd_args(self):
        parser = Parser(self, blnt.cache.commands)
        parser.run()



import traceback
from typing import Dict

import aiohttp_cors
from aiohttp import web as aio_web
from aiohttp.web_request import Request
from aiohttp.web_urldispatcher import Resource, ResourceRoute

from bolinette import console, blnt, web
from bolinette.exceptions import APIError, APIErrors, InternalError, InitError
from bolinette.utils.serializing import serialize


class BolinetteResources:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._aiohttp_resources: Dict[str, 'BolinetteResource'] = {}
        self._routes: Dict[str, Dict[web.HttpMethod, web.ControllerRoute]] = {}
        self.cors = aiohttp_cors.setup(self.context.app, defaults=self._setup_cors())

    @property
    def routes(self):
        for path, methods in self._routes.items():
            for method, route in methods.items():
                yield path, method, route

    def _setup_cors(self):
        try:
            conf = {}
            if 'cors' in self.context.env:
                conf = self.context.env['cors']
            if not isinstance(conf, dict):
                raise ValueError()
            defaults = {}
            for site, config in conf.items():
                if not isinstance(config, dict):
                    raise ValueError()
                defaults[site] = aiohttp_cors.ResourceOptions(
                    allow_credentials=config.get('allow_credentials', False),
                    expose_headers=config.get('expose_headers', ()),
                    allow_headers=config.get('allow_headers', ())
                )
            return defaults
        except ValueError:
            raise InitError("""
Invalid CORS config, you should have something like:

cors:
  "*":
    allow_credentials: true
    expose_headers: "*"
    allow_headers: "*"
  "http://client.example.org":
    allow_credentials: true
    expose_headers: "*"
    allow_headers: "*"
    max_age: 3600

See https://github.com/aio-libs/aiohttp-cors for detailed config options
""")

    def add_route(self, path: str, controller: 'web.Controller', route: 'web.ControllerRoute'):
        if path not in self._routes:
            self._routes[path] = {}
        self._routes[path][route.method] = route
        if path not in self._aiohttp_resources:
            self._aiohttp_resources[path] = BolinetteResource(self.cors.add(self.context.app.router.add_resource(path)))
        handler = RouteHandler(controller, route)
        self._aiohttp_resources[path].routes[route.method] = self.cors.add(
            self._aiohttp_resources[path].resource.add_route(route.method.http_verb, handler.__call__))


class BolinetteResource:
    def __init__(self, resource: Resource):
        self.resource = resource
        self.routes: Dict[web.HttpMethod, ResourceRoute] = {}


class RouteHandler:
    def __init__(self, controller: 'web.Controller', route: 'web.ControllerRoute'):
        self.controller = controller
        self.route = route

    async def __call__(self, request: Request):
        context: blnt.BolinetteContext = request.app['blnt']
        params = {
            'match': {},
            'query': {},
            'request': request
        }
        for key in request.match_info:
            params['match'][key] = request.match_info[key]
        for key in request.query:
            params['query'][key] = request.query[key]
        try:
            resp = await self.route.call_middleware_chain(request, params)
            return resp
        except (APIError, APIErrors) as ex:
            res = web.Response(context).from_exception(ex)
            if context.env['debug']:
                stack = traceback.format_exc()
                if isinstance(ex, InternalError):
                    console.error(stack)
                res.content['trace'] = stack.split('\n')
            serialized, mime = serialize(res.content, 'application/json')
            web_response = aio_web.Response(text=serialized, status=res.code, content_type=mime)
            return web_response



import re
from typing import Dict, List, Generator, Tuple, Any

from aiohttp import web as aio_web

from bolinette import blnt


class Topic:
    __blnt__: 'TopicMetadata' = None

    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self.__props__ = TopicProps(self)
        self._subscriptions: Dict[str, List[aio_web.WebSocketResponse]] = {}

    async def receive_subscription(self, channel: str, resp: aio_web.WebSocketResponse):
        if channel not in self._subscriptions:
            self._subscriptions[channel] = []
        self._subscriptions[channel].append(resp)

    async def validate_subscription(self, *args, **kwargs) -> bool:
        return True

    def _remove_closed_connections(self, channel: str) -> List[aio_web.WebSocketResponse]:
        subs = self._subscriptions.get(channel)
        if subs is None:
            return []
        self._subscriptions[channel] = list(filter(lambda c: not c.closed, subs))
        return self._subscriptions[channel]

    def subscriptions(self, channel: str) -> List[aio_web.WebSocketResponse]:
        return self._remove_closed_connections(channel)

    @staticmethod
    async def send(socket: aio_web.WebSocketResponse, data):
        await socket.send_json({'data': data})

    @staticmethod
    async def send_error(socket: aio_web.WebSocketResponse, error: str):
        await socket.send_json({'error': error})

    def __repr__(self):
        return f'<Topic {self.__blnt__.name}>'


class TopicMetadata:
    def __init__(self, name):
        self.name = name


class TopicProps:
    def __init__(self, topic: Topic):
        self.topic = topic

    def _get_attribute_of_type(self, attr_type):
        return ((name, attribute)
                for name, attribute in vars(self.topic.__class__).items()
                if isinstance(attribute, attr_type))

    def get_channels(self) -> Generator[Tuple[str, 'TopicChannel'], Any, None]:
        return self._get_attribute_of_type(TopicChannel)


class TopicChannel:
    def __init__(self, function, rule):
        self.function = function
        self.rule = rule
        self.re = re.compile(f'^{rule}$')

    def __repr__(self):
        return f'<Channel {self.rule}>'



import sys


class Console:
    def __init__(self, *, flush: bool = False):
        self.flush = flush

    def print(self, text: str = None, *, sep=' ', end='\n'):
        print(text or '', sep=sep, end=end, flush=self.flush)

    def error(self, text: str = None, *, sep=' ', end='\n'):
        print(text or '', file=sys.stderr, sep=sep, end=end, flush=self.flush)


console = Console()



from typing import List, Callable, Dict, Type

from bolinette import blnt, core, web
from bolinette.blnt.commands import Command


class BolinetteCache:
    def __init__(self):
        self.models: Dict[str, Type['core.Model']] = {}
        self.mixins: Dict[str, Type['core.Mixin']] = {}
        self.services: Dict[str, Type['core.Service']] = {}
        self.controllers: Dict[str, Type['web.Controller']] = {}
        self.middlewares: Dict[str, Type['web.Middleware']] = {}
        self.topics: Dict[str, Type['web.Topic']] = {}
        self.init_funcs: List[Callable[['blnt.BolinetteContext'], None]] = []
        self.commands: Dict[str, Command] = {}
        self.seeders = []


cache = BolinetteCache()



from typing import Union, Literal

from bolinette import types


class Reference:
    def __init__(self, model_name: str, column_name: str):
        self.model_name = model_name
        self.column_name = column_name

    def __repr__(self):
        return f'<Reference -> {self.model_name}.{self.column_name}'


class Column:
    def __init__(self, data_type: 'types.db.DataType', *, reference: Reference = None, primary_key: bool = False,
                 nullable: bool = True, unique: bool = False, model_id: bool = False, default=None):
        self.type = data_type
        self.reference = reference
        self.primary_key = primary_key
        self.nullable = nullable
        self.unique = unique
        self.model_id = model_id
        self.name = None
        self.default = default

    def __repr__(self):
        return f'<Column {self.name}: {repr(self.type)}>'


class Backref:
    def __init__(self, key: str, *, lazy: bool = True):
        self.key = key
        self.lazy = lazy

    def __repr__(self):
        return f'<Backref <- {self.key}' + (' lazy ' if self.lazy else '') + '>'


class Relationship:
    def __init__(self, model_name: str, *, backref: Backref = None, foreign_key: Column = None,
                 lazy: Union[bool, Literal['subquery']] = False, secondary: str = None, remote_side: Column = None):
        self.model_name = model_name
        self.foreign_key = foreign_key
        self.backref = backref
        self.secondary = secondary
        self.remote_side = remote_side
        self.lazy = lazy
        self.name = None

    def __repr__(self):
        return f'<Relationship {self.name} -> {self.model_name}' + (' lazy ' if self.lazy else '') + '>'



from aiohttp import web as aio_web

from bolinette import web
from bolinette.decorators import controller, get
from bolinette.defaults.services import FileService


@controller('file', '/file', middlewares=['auth'])
class FileController(web.Controller):
    @property
    def file_service(self) -> FileService:
        return self.context.service('file')

    def default_routes(self):
        return [
            self.defaults.get_one(key='key')
        ]

    @get('/{key}/download')
    async def download_file(self, match):
        file = await self.service.get_first_by('key', match['key'])
        headers = {
            'Content-disposition': f'attachment; filename={file.name}'
        }
        return aio_web.Response(body=self.file_service.file_sender(file.key),
                                headers=headers, content_type=file.mime)



from bolinette import blnt
from bolinette.decorators import command


@command('init_docs', 'Initialize the OpenAPI documentation')
async def init_docs(context: 'blnt.BolinetteContext'):
    context.docs.build()



from typing import List, Callable

from bolinette.blnt.commands import Argument


class Command:
    def __init__(self, name: str, func: Callable, summary: str = None, args: List[Argument] = None):
        self.name = name
        self.func = func
        self.summary = summary
        self.args = args or []



from typing import Dict

from bolinette import types


class Mixin:
    @staticmethod
    def columns() -> Dict[str, 'types.defs.Column']:
        pass

    @staticmethod
    def relationships(model_cls) -> Dict[str, 'types.defs.Relationship']:
        pass



from typing import Dict

from bolinette import core, blnt, mapping, types
from bolinette.exceptions import InternalError
from bolinette.utils.functions import getattr_, hasattr_, invoke


class Mapper:
    def __init__(self, context: 'blnt.BolinetteContext'):
        self.context = context
        self._payloads: Dict[str, Dict[str, mapping.Definition]] = {}
        self._responses: Dict[str, Dict[str, mapping.Definition]] = {}

    @staticmethod
    def _get_def(collection, model_name, key) -> 'mapping.Definition':
        m = collection.get(model_name)
        if m is None:
            raise InternalError(f'mapping.unknown_model:{model_name}')
        d = m.get(key)
        if d is None:
            raise InternalError(f'mapping.unknown_definition:{key}')
        return d

    def payload(self, model_name: str, key: str):
        return self._get_def(self._payloads, model_name, key)

    @property
    def payloads(self):
        for model_name in self._payloads:
            for key in self._payloads[model_name]:
                yield model_name, key, self._payloads[model_name][key]

    def response(self, model_name: str, key: str):
        return self._get_def(self._responses, model_name, key)

    @property
    def responses(self):
        for model_name in self._responses:
            for key in self._responses[model_name]:
                yield model_name, key, self._responses[model_name][key]

    def register(self, model_name: str, model: 'core.Model'):
        def create_defs(collection, params):
            if params is None:
                return
            for param in params:
                if isinstance(param, list):
                    model_key = 'default'
                    payload = param
                else:
                    model_key, payload = param
                definition = mapping.Definition(model_name, model_key)
                for field in payload:
                    definition.fields.append(field)
                if definition.model_name not in collection:
                    collection[definition.model_name] = {}
                collection[definition.model_name][definition.model_key] = definition

        create_defs(self._payloads, model.payloads())
        create_defs(self._responses, model.responses())

    def marshall(self, definition, entity, *, skip_none=False, as_list=False, use_foreign_key=False):
        if not entity:
            return None
        if as_list:
            return [self.marshall(definition, e, skip_none=skip_none, as_list=False, use_foreign_key=use_foreign_key)
                    for e in entity]
        values = {}
        for field in definition.fields:
            self._marshall_object(values, field, entity, skip_none, use_foreign_key)
        return values

    def _marshall_object(self, values, field, entity, skip_none: bool, use_foreign_key: bool):
        if isinstance(field, mapping.Field):
            self._marshall_field(values, field, entity, skip_none)
        elif isinstance(field, mapping.Reference) and use_foreign_key:
            values[field.foreign_key] = getattr_(entity, field.foreign_key, None)
        elif isinstance(field, mapping.Definition):
            self._marshall_definition(values, field, entity, skip_none, use_foreign_key)
        elif isinstance(field, mapping.List):
            self._marshall_list(values, field, entity, skip_none, use_foreign_key)

    @staticmethod
    def _marshall_field(values, field: 'mapping.Field', entity, skip_none: bool):
        if field.function is not None:
            value = field.function(entity)
        else:
            value = getattr_(entity, field.key, None)
        if field.formatting is not None:
            value = field.formatting(value)
        if not skip_none or value is not None:
            values[field.name] = value

    def _marshall_definition(self, values, definition: 'mapping.Definition', entity,
                             skip_none: bool, use_foreign_key: bool):
        d = self.response(definition.model_name, definition.model_key)
        attr = None
        if definition.function and callable(definition.function):
            attr = definition.function(entity)
        elif hasattr_(entity, definition.name):
            attr = getattr_(entity, definition.name, None)
        values[definition.name] = self.marshall(d, attr, skip_none=skip_none, as_list=False,
                                                use_foreign_key=use_foreign_key)

    def _marshall_list(self, values, field: 'mapping.List', entity, skip_none: bool, use_foreign_key: bool):
        if field.function and callable(field.function):
            e_list = invoke(field.function, entity)
        else:
            e_list = getattr_(entity, field.name, None)
        elem = field.element
        if isinstance(elem, types.db.DataType):
            values[field.name] = [e for e in e_list]
        elif isinstance(elem, mapping.Definition):
            d = self.response(elem.model_name, elem.model_key)
            values[field.name] = self.marshall(d, e_list, skip_none=skip_none, as_list=True,
                                               use_foreign_key=use_foreign_key)



import os

from bolinette.utils import files, paths
from setuptools import setup, find_packages

from bolinette import blnt


context = blnt.BolinetteContext(paths.dirname(__file__), None)


def project_packages(module):
    return [m for m in find_packages() if m.startswith(module)]


setup(
    name='Bolinette',
    packages=project_packages('bolinette'),
    include_package_data=True,
    version=os.environ['BLNT_RELEASE_VERSION'],
    license='MIT',
    description='Bolinette, a web framework built on top of Flask',
    long_description=files.read_file(context.root_path('README.md')),
    long_description_content_type='text/markdown',
    author='Pierre Chat',
    author_email='pierrechat@outlook.com',
    url='https://github.com/TheCaptainCat/bolinette',
    keywords=['Flask', 'Bolinette', 'Web', 'Framework'],
    install_requires=files.read_requirements(context.root_path()),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    setup_requires=[
        'wheel'
    ]
)



import sqlalchemy
from sqlalchemy import orm as sqlalchemy_orm

from bolinette import blnt, core
from bolinette.decorators import init_func
from bolinette.exceptions import InitError


@init_func
def init_relational_models(context: blnt.BolinetteContext):
    models = {}
    for model_name, model_cls in blnt.cache.models.items():
        db_key = model_cls.__blnt__.database
        if db_key in context.db:
            if not context.db[db_key].relational:
                continue
            database = context.db[db_key]
        else:
            raise InitError(f'Undefined "{db_key}" database for model "{model_name}"')
        models[model_name] = model_cls(database)
    orm_tables = {}
    orm_cols = {}
    for model_name, model in models.items():
        orm_cols[model_name] = {}
        for att_name, attribute in model.__props__.get_columns():
            attribute.name = att_name
            ref = None
            if attribute.reference:
                ref = sqlalchemy.ForeignKey(f'{attribute.reference.model_name}.{attribute.reference.column_name}')
            orm_cols[model_name][att_name] = sqlalchemy.Column(
                att_name, attribute.type.sqlalchemy_type, ref, default=attribute.default, index=attribute.model_id,
                primary_key=attribute.primary_key, nullable=attribute.nullable, unique=attribute.unique)
        orm_tables[model_name] = sqlalchemy.Table(model_name,
                                                  model.__props__.database.base.metadata,
                                                  *(orm_cols[model_name].values()))

    for model_name, model in models.items():
        orm_defs = {}
        for att_name, attribute in model.__props__.get_relationships():
            kwargs = {}
            attribute.name = att_name
            if attribute.backref:
                kwargs['backref'] = sqlalchemy_orm.backref(attribute.backref.key, lazy=attribute.backref.lazy)
            if attribute.foreign_key:
                kwargs['foreign_keys'] = orm_cols[model_name][attribute.foreign_key.name]
            if attribute.remote_side:
                kwargs['remote_side'] = orm_cols[model_name][attribute.remote_side.name]
            if attribute.secondary:
                kwargs['secondary'] = orm_tables[attribute.secondary]
            orm_defs[att_name] = sqlalchemy_orm.relationship(attribute.model_name,  lazy=attribute.lazy, **kwargs)

        orm_defs['__table__'] = orm_tables[model_name]
        orm_model = type(model_name, (model.__props__.database.base,), orm_defs)

        for att_name, attribute in model.__props__.get_properties():
            setattr(orm_model, att_name, property(attribute.function))

        context.add_model(model_name, model)
        context.add_table(model_name, orm_model)


@init_func
def init_collection_models(context: blnt.BolinetteContext):
    models = {}
    for model_name, model_cls in blnt.cache.models.items():
        db_key = model_cls.__blnt__.database
        if db_key in context.db:
            if context.db[db_key].relational:
                continue
            database = context.db[db_key]
        else:
            raise InitError(f'Undefined "{db_key}" database for model "{model_name}"')
        models[model_name] = model_cls(database)
    for model_name, model in models.items():
        for att_name, attribute in model.__props__.get_columns():
            attribute.name = att_name
        context.add_model(model_name, model)


@init_func
async def init_databases(context: blnt.BolinetteContext):
    await context.db.create_all()


@init_func
def init_repositories(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        context.add_repo(model_name, core.Repository(model_name, model, context))


@init_func
def init_mappings(context: blnt.BolinetteContext):
    for model_name, model in context.models:
        context.mapper.register(model_name, model)


@init_func
def init_services(context: blnt.BolinetteContext):
    for service_name, service_cls in blnt.cache.services.items():
        context.add_service(service_name, service_cls(context))


@init_func
def init_controllers(context: blnt.BolinetteContext):
    for controller_name, controller_cls in blnt.cache.controllers.items():
        controller = controller_cls(context)
        for _, route in controller.__props__.get_routes():
            route.controller = controller
            route.setup(controller)
        for route in controller.default_routes():
            route.controller = controller
            route.setup(controller)
        context.add_controller(controller_name, controller)


@init_func
def init_topics(context: blnt.BolinetteContext):
    context.sockets.init_socket_handler()
    for topic_name, topic_cls in blnt.cache.topics.items():
        topic = topic_cls(context)
        context.sockets.add_topic(topic_name, topic)
        for channel_name, channel in topic.__props__.get_channels():
            context.sockets.add_channel(topic_name, channel)



from bolinette.blnt.logging import Logger
from bolinette.blnt.cache import cache
from bolinette.blnt.environment import init, Environment
from bolinette.blnt.jwt import JWT
from bolinette.blnt.validation import Validator
from bolinette.blnt.context import BolinetteContext
from bolinette.blnt.transaction import Transaction



import random
import string
from datetime import datetime

from bolinette import blnt, web
from bolinette.decorators import controller, get, post, patch, delete
from bolinette.defaults.services import UserService, RoleService
from bolinette.exceptions import ForbiddenError


@controller('user', '/user')
class UserController(web.Controller):
    @property
    def user_service(self) -> UserService:
        return self.context.service('user')

    @property
    def role_service(self) -> RoleService:
        return self.context.service('role')

    def default_routes(self):
        return [
            self.defaults.get_all('private', middlewares=['auth|roles=admin']),
            self.defaults.get_one('private', prefix='/u', key='username', middlewares=['auth|roles=admin'])
        ]

    def _create_tokens(self, resp, user, *, set_access, set_refresh, fresh):
        now = datetime.utcnow()
        if set_access:
            access_token = self.context.jwt.create_access_token(now, user.username, fresh=fresh)
            resp.cookies.append(web.Cookie('access_token', access_token,
                                           expires=self.context.jwt.access_token_expires(now), path='/'))
        if set_refresh:
            refresh_token = self.context.jwt.create_refresh_token(now, user.username)
            resp.cookies.append(web.Cookie('refresh_token', refresh_token,
                                           expires=self.context.jwt.refresh_token_expires(now),
                                           path='/api/user/refresh'))

    @get('/info', returns=web.Returns('user', 'private'), middlewares=['auth'])
    async def info(self, current_user):
        """
        Gets current user's details
        """
        return current_user

    @post('/login', expects=web.Expects('user', 'login'), returns=web.Returns('user', 'private'))
    async def login(self, payload):
        """
        Logs the user in, setting the JWT inside the cookies
        """
        username = payload['username']
        password = payload['password']
        user = await self.user_service.get_by_username(username, safe=True)
        if user is not None:
            if self.user_service.check_password(user, password):
                resp = self.response.ok(messages='user.login.success', data=user)
                self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
                return resp
        return self.response.unauthorized(messages='user.login.wrong_credentials')

    @post('/logout')
    async def logout(self):
        """
        Deletes authentication cookies, effectively logging the user off
        """
        resp = self.response.ok(messages='user.logout.success')
        resp.cookies.append(web.Cookie('access_token', None, delete=True, path='/'))
        resp.cookies.append(web.Cookie('refresh_token', None, delete=True, path='/api/user/refresh'))
        return resp

    @post('/token/refresh', middlewares=['auth'])
    async def refresh(self, current_user):
        """
        Creates a new fresh JWT
        """
        resp = self.response.ok(messages='user.token.refreshed')
        self._create_tokens(resp, current_user, set_access=True, set_refresh=False, fresh=False)
        return resp

    @post('/register', expects=web.Expects('user', 'register'), returns=web.Returns('user', 'private'))
    async def register(self, payload):
        """
        Creates a new user from given information

        If the init parameter ADMIN_REGISTER_ONLY is true, only admin users can create new accounts and
        the route will send back a 403 FORBIDDEN.

        -response 201 returns: The created user
        -response 403: If the route is disabled
        """
        if blnt.init.get('ADMIN_REGISTER_ONLY', True):
            raise ForbiddenError('global.register.admin_only')
        user = await self.user_service.create(payload)
        resp = self.response.created(messages='user.registered', data=user)
        self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
        return resp

    @post('/register/admin', expects=web.Expects('user', 'admin_register'),
          returns=web.Returns('user', 'private'), middlewares=['auth|roles=admin'])
    async def admin_register(self, payload):
        """
        Creates a new user from given information and a random password

        Admin only route.
        """
        payload['password'] = ''.join(random.choices(string.ascii_lowercase, k=32))
        user = await self.user_service.create(payload)
        return self.response.created(messages='user.registered', data=user)

    @patch('/me', expects=web.Expects('user', 'register', patch=True), returns=web.Returns('user', 'private'),
           middlewares=['auth|fresh'])
    async def update_user(self, payload, current_user):
        """
        Updates current user information from payload

        Requires a fresh JWT.
        """
        user = await self.user_service.patch(current_user, payload)
        resp = self.response.ok(messages='user.updated', data=user)
        self._create_tokens(resp, user, set_access=True, set_refresh=True, fresh=True)
        return resp

    @post('/{username}/roles', expects=web.Expects('role'), returns=web.Returns('user', 'private'),
          middlewares=['auth|roles=admin'])
    async def add_user_role(self, match, payload):
        """
        Adds a role to the user identified by username param

        Admin only route.
        """
        user = await self.user_service.get_by_username(match['username'])
        role = await self.role_service.get_by_name(payload['name'])
        await self.user_service.add_role(user, role)
        return self.response.created(messages=f'user.roles.added:{user.username}:{role.name}', data=user)

    @delete('/{username}/roles/{role}', returns=web.Returns('user', 'private'), middlewares=['auth|roles=admin'])
    async def delete_user_role(self, match, current_user):
        """
        Removes a role from the user identified by username parameter

        Admin only route.
        """
        user = await self.user_service.get_by_username(match['username'])
        role = await self.role_service.get_by_name(match['role'])
        await self.user_service.remove_role(current_user, user, role)
        return self.response.ok(messages=f'user.roles.removed:{user.username}:{role.name}', data=user)

    @post('/picture', returns=web.Returns('user', 'private'), middlewares=['auth'])
    async def upload_profile_picture(self, current_user, payload):
        """
        Uploads a profile picture for the current user
        """
        picture = payload['file']
        user = await self.user_service.save_profile_picture(current_user, picture)
        return self.response.ok(messages=f'user.picture.uploaded', data=user)



import getpass

import bolinette
from bolinette import Console
from bolinette.blnt import Transaction
from bolinette.decorators import command
from bolinette.exceptions import ParamConflictError, EntityNotFoundError, APIError, APIErrors


@command('create_user', 'Add a user to the database')
@command.argument('argument', 'username', summary='The new user\'s username')
@command.argument('argument', 'email', summary='The new user\'s email')
@command.argument('option', 'roles', flag='r', summary='The user\'s roles, comma separated')
async def create_user(blnt: 'bolinette.Bolinette', username: str, email: str, roles: str = None):
    console = Console()
    user_service = blnt.context.service('user')
    role_service = blnt.context.service('role')
    while True:
        password = getpass.getpass('Choose password: ')
        password2 = getpass.getpass('Confirm password: ')
        if password == password2:
            break
        console.error('Passwords don\'t match')
    async with Transaction(blnt.context, print_error=False, propagate_error=False):
        user_roles = []
        if roles is not None:
            for role_name in [r.strip() for r in roles.split(',')]:
                try:
                    user_roles.append(await role_service.get_by_name(role_name))
                except EntityNotFoundError:
                    console.error(f'Role "{role_name}" does not exist')
                    exit(1)
        try:
            user = await user_service.create({
                'username': username,
                'password': password,
                'email': email
            })
            for role in user_roles:
                user.roles.append(role)
        except (APIError, APIErrors) as ex:
            if isinstance(ex, APIError):
                errors = [ex]
            else:
                errors = ex.errors
            for err in [err for err in errors if isinstance(err, ParamConflictError)]:
                console.error(f'Conflict: {err.message.split(":")[1]} already exists')
            exit(1)
